/********************************************************************
* The following program permits modifications to the NT virtual 
*   memory settings from the command line.  It does the same thing
*   as going thru the control panel, system, performance, change
*   button.  It only changes virtual memory settings, it does not
*   modify the max registry size setting.
*
* This is possible because the virtual memory settings are kept in a
*   REG_MULTI_SZ value called PagingFiles in the HKEY_LOCAL_MACHINE\
*   SYSTEM\CurrentControlSet\Control\Session Manager\
*   Memory Management key of the registry.
*
* 09 Apr. 1998 - Initial implementation.  Registry open and update
*                  funcitonality implemented.
*
* 13 Apr. 1998 - Added a single command line option to increase the
*                  initial and max size of the first pagefile.
*
* 21 Apr. 1998 - Began to add the real command line parsing code.
*                  Added the opts array, and the code to fill it with
*                  the argv offsets of switches.
*              - Added the CMD and OPT macros.
*              - Added the sw_ variables to hold the values passed
*                  with the modifiers.
*
* 22 Apr. 1998 - Finished the code for the ADD command.
*              - Added the code for the SUB command.
*              - Added the update flag, so the code for updating the
*                  registry key from the pagefile array is in only
*                  place.
*              - Added a deleted flag to the pfile structure.  If 
*                  this flag is 0 then the structure gets written
*                  out during the registry update.  Otherwise, the
*                  structure gets skipped during the update.
*
* 23 Apr. 1998 - Removed the MOVE command.  A move can actually be 
*                  performed with a DELETE, and then a NEW.
*              - Created the function, matchpage2drv, to locate the
*                  pagefile on a specific drive.
*              - In keydata2struct, added code to convert the path
*                  to uppercase.  This makes it easier to search thru
*                  the array with just a drive letter.
*              - Added the /I and /X modifiers for the Query command.
*              - Added code for the DELETE command.
*              - Added code for the NEW command.
*              - Changed the code on the ADD and SUB commands so that
*                  the /I switch allows an increase of just the 
*                  initial size, and the /X switch allows an increase
*                  to just the maximum size, while the /A switch 
*                  increases both by an equal amount.  /I and /X have
*                  precedence over /A.
*              - Changed the matchpage2drv function to return -1 if
*                  a pagefile was not found on the spec'f drive.
*              - Got rid of the VERBOSE modifier.
*              - Added code for the QUERY cmd.  
*                  /D[A-Z] by itself returns 1 if there is a pagefile
*                    on the drive, otherwise it returns 0.
*                  /D /I returns the initial size of the pagefile on
*                    the spec'd drive.  0 if there is no pagefile on
*                    the drive.
*                  /D /X returns the maximum size of the pagefile on
*                    the spec'd drive.  0 if there is no pagefile on
*                    the drive.
*                  If both the /I and the /X modifiers are passed, /I
*                    has precedence.
*              - Application is now complete except for returning 
*                  error conditions.
*              - Error codes and processing works.
*
* 01 May. 1998 - Tested under Nt 5.0 Beta 1.  Worked no problem.
*
* Return codes - >= 16,777,216 is always an error (0x01000000)
*                0x01000000 - Usage message.
*                0x02000000 - Invalid or no command
*                0x03000000 - Opts didn't get allocated
*                0x04000000 - Error reading registry value
*                0x05000000 - Error opeing the registry
*
* Compile: cl /nologo /W3 /D "WIN32" /D "CONSOLE"  /Zp1 /c batchvmm.c
* Link   : link /NOLOGO /MACHINE:IX86 /SUBSYSTEM:CONSOLE 
*               /OUT:batchvmm.exe batchvmm.obj advapi32.lib
********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <winnt.h>
#include <winbase.h>                             
#include <winreg.h>               /* Registry API */

#define CMD_QUERY   1             /* Query command spec'd */
#define CMD_ADD     2             /* Add command spec'd */
#define CMD_SUB     4             /* Sub command spec'd */
#define CMD_NEW     8             /* New command spec'd */
#define CMD_DELETE 16             /* Delete command spec'd */
#define CMD_NOCMD  0xFFFFFFFF     /* Invlaid command spec'd */
 
#define OPT_AMT     1             /* Amount switch passed */
#define OPT_PAG     2             /* Pagefile switch passed */
#define OPT_DRV     4             /* Drive switch passed */
#define OPT_INIT    8             /* Initial size switch passed */
#define OPT_MAX    16             /* Maximum size switch passed */
#define OPT_NOOPT   0             /* No switches passed */

#define VMM_KEY "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Memory Management"
#define VMM_VAL "PagingFiles"

#define PG_PATH "X:\\"            /* Pagefile path - used in NEW */
#define PG_NAME "pagefile.sys"    /* Pagefile name - used in NEW */

struct pfile {                    /* Defines one line value */
  unsigned char *path;            /* Pagefile location */
  unsigned char *name;            /* Pagefile name */
  unsigned int   init;            /* Initial size in MB */
  unsigned int   max;             /* Maximum size in MB */
  unsigned int   deleted;         /* 0=use, 1=don't use */
};

signed int matchpage2drv(int n_pgs,char drv,struct pfile *pg_fls) {
  int ctr;                        /* FOR loop counter */
  int rtc;                        /* Return code */

  for(ctr=0; ctr<n_pgs; ctr++) {
    if(pg_fls[ctr].path[0]==drv) {
      rtc=ctr;
      break;
    }
    else
      continue;
  }

  if(ctr==n_pgs)
    rtc=-1;

  return(rtc);
}

void struct2wbuffer(int n_pgs,unsigned char *keydata,
                    struct pfile *pg_fls) {
  unsigned char  init_buf[18];    /* Init field string conv buffer */
  unsigned char  max_buf[18];     /* Max field string conv buffer */
  unsigned char *cur;             /* Current location in data */
  int            ctr;             /* FOR loop counter */

  cur=keydata;
  for(ctr=0; ctr<n_pgs; ctr++) {
    if(pg_fls[ctr].deleted==0) {
      strcpy(cur,pg_fls[ctr].path);
      cur+=strlen(pg_fls[ctr].path);
      strcpy(cur,pg_fls[ctr].name);
      cur+=strlen(pg_fls[ctr].name);
      *cur=' ';
      cur++;
      memset(init_buf,'\0',19);
      itoa(pg_fls[ctr].init,init_buf,10);
      strcpy(cur,init_buf);
      cur+=strlen(init_buf);
      *cur=' ';
      cur++;
      memset(max_buf,'\0',19);
      itoa(pg_fls[ctr].max,max_buf,10);
      strcpy(cur,max_buf);
      cur+=strlen(max_buf);
      cur+=1;
    }
    else
      continue;
  }
}

struct pfile *keydata2struct(int *n_pgs, 
                             unsigned char *keydata, 
                             unsigned long *w_bf_sz) {
  unsigned char *tmp0;            /* Scratch parsing variable */
  unsigned char *tmp1;            /* Scratch parsing variable */
  unsigned char *cur;             /* Current location in data */
  int            ctr;             /* FOR loop counter */
  struct pfile  *pg_fls;          /* Array of page file data */

  pg_fls=(struct pfile *)calloc(*n_pgs,sizeof(struct pfile));

  cur=keydata;                    /* Return to beginning */
  for(ctr=0; ctr<*n_pgs; ctr++) { /* Initialize array */

    tmp0=strrchr(cur,'\\');       /* Find last \ */
    pg_fls[ctr].path=(unsigned char *)calloc((tmp0-cur)+2,
                                            sizeof(unsigned char));
    strncpy(pg_fls[ctr].path,cur,(tmp0-cur)+1);
    strupr(pg_fls[ctr].path);
    tmp1=tmp0;
    tmp1++;
    
    tmp0=strchr(tmp1,' ');
    pg_fls[ctr].name=(unsigned char *)calloc((tmp0-tmp1)+1,
                                            sizeof(unsigned char));
    strncpy(pg_fls[ctr].name,tmp1,tmp0-tmp1);
    tmp1=tmp0;
    tmp1++;

    pg_fls[ctr].init=atoi(tmp1);
    tmp0=strchr(tmp1,' ');
    if(tmp0!=NULL) {
      tmp1=tmp0;
      tmp1++;
      pg_fls[ctr].max=atoi(tmp1);
    }
    else {
      pg_fls[ctr].max=pg_fls[ctr].init;
    }

    cur=strchr(cur,'\0');
    cur++;

    /****************************************************************
    * Total the size of the data buffer needed to update the VALUE.
    ****************************************************************/
    *w_bf_sz+=strlen(pg_fls[ctr].path);
    *w_bf_sz+=strlen(pg_fls[ctr].name);
    *w_bf_sz+=19;                 /* Space + number */
    *w_bf_sz+=19;                 /* Space + number */
    *w_bf_sz+=2;                  /* Terminating null(s) */
  }  
  return(pg_fls);
}

void print_keydata(unsigned char *keydata) {
  unsigned char *cur;             /* Current location in data */

  cur=keydata;                    /* Loop thru value entries */
  while(*cur!='\0') {             /* While current isn't NULL */
    printf("%s\n",cur);           /* Print the current string */
    cur=strchr(cur,'\0');         /* Goto end of this string */
    cur++;                        /* Goto start of next string */
  }
}

int count_pagefiles(unsigned char *keydata) {
  int            num_page_files=0;/* Number of pagefiles */
  unsigned char *cur;             /* Current location in data */

  cur=keydata;                    /* Loop thru value entries */
  while(*cur!='\0') {             /* While current isn't NULL */
    num_page_files++;             /* Increment number of files */
    cur=strchr(cur,'\0');         /* Goto end of this string */
    cur++;                        /* Goto start of next string */
  }
  return(num_page_files);         /* Return the result */
}

int main(int argc, char *argv[]) {
  unsigned char *r_data;          /* Current registry value data */
  unsigned char *w_data;          /* New registry value data */
  char           sw_drive=0;      /* Drive modifier value */
  int            num_pg=0;        /* Number of paging files */
  int            ctr;             /* FOR loop counter */
  int           *opts;            /* Array of command line switches */
  int            errorlevel;      /* MAIN return code */
  int            command=0;       /* What command was argv[1] */
  int            switches=0;      /* What switches were spec'd */
  int            sw_amount=0;     /* Amount modifier value */
  int            sw_initial=0;    /* Initial modifier value */
  int            sw_maximum=0;    /* Maximum modifier value */
  int            update=0;        /* 1=write out change, 0=don't */
  signed int     sw_pagefile=0;   /* Pagefile modifier value */
  unsigned long  type;            /* Type of registry value */
  unsigned long  r_data_size=0;   /* Bytes needed for read */
  unsigned long  w_data_size=0;   /* Bytes needed for write */
  unsigned long  rtc;             /* Return code */
  struct pfile  *files;           /* Array of page file data */
  struct pfile  *old;             /* Used only with NEW command */
  HKEY           key;             /* Registry key handle */

  if(argc<2) {
    printf("Usage: BATCHVMM COMMAND Modifiers\n");
    printf("\n");
    printf("Where COMMAND is one of the following:\n");
    printf("\tQUERY  - Query Virtual Memory Settings\n");
    printf("\tADD    - Add AMOUNT to PAGEFILE or DRIVE\n");
    printf("\tSUB    - Subtract AMOUNT from PAGEFILE or DRIVE\n");
    printf("\tNEW    - Create a new pagefile on DRIVE\n");
    printf("\tDELETE - Delete PAGEFILE or DRIVE\n");
    printf("\n");
    printf("Modifiers are:\n");
    printf("\t/A     - Amount   /Annn\n");
    printf("\t/P     - Pagefile /Pnnn\n");
    printf("\t/D     - Drive    /D[A-Z]\n");
    printf("\t/I     - Initial size /Innn\n");
    printf("\t/X     - Maximum size /Xnnn\n");
    errorlevel=0x01000000;
    goto EXIT;
  }

  strupr(argv[1]);                /* Make the command upper case */
  
  /******************************************************************
  * Check argv[1] against all possible commands.  If argv[1] does not
  *   match any of the defined commands, set command to CMD_NOCMD.
  ******************************************************************/
  if(strcmp(argv[1],"QUERY")==0)
    command=CMD_QUERY;
  else if(strcmp(argv[1],"ADD")==0)
    command=CMD_ADD;
  else if(strcmp(argv[1],"SUB")==0)
    command=CMD_SUB;
  else if(strcmp(argv[1],"NEW")==0)
    command=CMD_NEW;
  else if(strcmp(argv[1],"DELETE")==0)
    command=CMD_DELETE;
  else
    command=CMD_NOCMD;

  if(command==CMD_NOCMD) {        /* If invalid or no command */
    errorlevel=0x02000000;        /* Return code is 0x02000000 */
    goto EXIT;                    /* Terminate with return code */
  }

  /******************************************************************
  * Allocate the array of argv offsets and initialize.  Either the -
  *   or the / is a valid option switch.
  ******************************************************************/
  opts=(int *)calloc(argc,sizeof(int));
  if(opts==NULL) {                /* If opts didn't get allocated */
    errorlevel=0x03000000;        /* Return code is 0x03000000 */
    goto EXIT;                    /* Terminate with return code */
  }
  for(ctr=2; ctr<argc; ctr++) {
    if(argv[ctr][0]=='-' || argv[ctr][0]=='/')
      opts[ctr-2]=ctr;
    else
      continue;
  }

  /******************************************************************
  * Process the modifiers.
  ******************************************************************/
  ctr=0;
  while(opts[ctr]!=0) {
    switch(argv[opts[ctr]][1]) {
      case 'a' :
      case 'A' : switches|=OPT_AMT;
                 sw_amount=atoi(&argv[opts[ctr]][2]);
                 printf("Amount = %d\n",sw_amount);
                 break;
      case 'd' :
      case 'D' : switches|=OPT_DRV;
                 sw_drive=argv[opts[ctr]][2];
                 if(sw_drive<122 && sw_drive>97)
                   sw_drive-=32;
                 printf("Drive = %c\n",sw_drive);
                 break;
      case 'i' :
      case 'I' : switches|=OPT_INIT;
                 if(command!=CMD_QUERY) {
                   sw_initial=atoi(&argv[opts[ctr]][2]);
                   printf("Initial = %d\n",sw_initial);
                 }
                 else
                   printf("Query modifier - Initial\n");
                 break;
      case 'p' :
      case 'P' : switches|=OPT_PAG;
                 sw_pagefile=atoi(&argv[opts[ctr]][2]);
                 printf("Pagefile = %d\n",sw_pagefile);
                 break;
      case 'x' :
      case 'X' : switches|=OPT_MAX;
                 if(command!=CMD_QUERY) {
                   sw_maximum=atoi(&argv[opts[ctr]][2]);
                   printf("Maximum = %d\n",sw_maximum);
                 }
                 else
                   printf("Query modifier - Maximum\n");
                 break;
    }
    ctr++;                        /* Increment ctr */
  }

  /******************************************************************
  * Open the registry, and retrieve the value of the PagingFiles
  *   value.  Then, count the number of pagefiles, and load the
  *   data into an array of pfile structs.
  ******************************************************************/
  rtc=RegOpenKeyEx(HKEY_LOCAL_MACHINE,VMM_KEY,0,KEY_ALL_ACCESS,&key);
  if(rtc!=ERROR_SUCCESS) {        /* If error opening registry */
    errorlevel=0x05000000;        /* Return code is 0x05000000 */
    goto EXIT;                    /* Terminate with return code */
  }

  rtc=RegQueryValueEx(key,VMM_VAL,NULL,&type,r_data,&r_data_size);
  if(rtc==ERROR_MORE_DATA) {
    r_data=(unsigned char *)calloc(r_data_size,
                                   sizeof(unsigned char));
    rtc=RegQueryValueEx(key,VMM_VAL,NULL,&type,r_data,&r_data_size);
  }
  if(rtc!=ERROR_SUCCESS) {        /* If error reading value */
    errorlevel=0x04000000;        /* Return code is 0x04000000 */
    goto EXIT;                    /* Terminate with return code */
  }

  num_pg=count_pagefiles(r_data); /* Count pagefiles */
  files=keydata2struct(&num_pg,r_data,&w_data_size);

  /******************************************************************
  * Based on the command specified in argv[1], perform the
  *   appropriate tasks.
  ******************************************************************/
  switch(command) {
    case CMD_QUERY  : if(!(switches&OPT_DRV))
                        break;
                      if(switches==OPT_DRV) {
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,
                                                  files);
                        if(sw_pagefile==-1)
                          errorlevel=0;
                        else
                          errorlevel=1;
                        break;
                      }
                      if(switches&OPT_INIT) {
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,
                                                  files);
                        if(sw_pagefile==-1)
                          errorlevel=0;
                        else
                          errorlevel=files[sw_pagefile].init;
                        break;
                      }
                      if(switches&OPT_MAX) {
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,
                                                  files);
                        if(sw_pagefile==-1)
                          errorlevel=0;
                        else
                          errorlevel=files[sw_pagefile].max;
                        break;
                      }
                      break;
    case CMD_ADD    : if((!(switches&OPT_AMT))&&
                         (!(switches&OPT_INIT))&&
                         (!(switches&OPT_MAX))) {
                        break;
                      }
                      if((switches&OPT_DRV)&&(!(switches&OPT_PAG)))
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,
                                                  files);
                      if(sw_pagefile==-1)
                        break;
                      if(switches&OPT_INIT)
                        files[sw_pagefile].init+=sw_initial;
                      if(switches&OPT_MAX)
                        files[sw_pagefile].max+=sw_maximum;
                      if((!(switches&OPT_INIT))&&
                         (!(switches&OPT_MAX))) {
                        files[sw_pagefile].init+=sw_amount;
                        files[sw_pagefile].max+=sw_amount;
                      }
                      update=1;
                      break;
    case CMD_SUB    : if((!(switches&OPT_AMT))&&
                         (!(switches&OPT_INIT))&&
                         (!(switches&OPT_MAX))) {
                        break;
                      }
                      if((switches&OPT_DRV)&&(!(switches&OPT_PAG)))
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,
                                                  files);
                      if(sw_pagefile==-1)
                        break;
                      if(switches&OPT_INIT)
                        files[sw_pagefile].init-=sw_initial;
                      if(switches&OPT_MAX)
                        files[sw_pagefile].max-=sw_maximum;
                      if((!(switches&OPT_INIT))&&
                         (!(switches&OPT_MAX))) {
                        files[sw_pagefile].init-=sw_amount;
                        files[sw_pagefile].max-=sw_amount;
                      }
                      update=1;
                      break;
    case CMD_NEW    : old=files;
                      num_pg++;
                      files=(struct pfile *)calloc(num_pg,
                                              sizeof(struct pfile));
                      memcpy(files,old,
                             sizeof(struct pfile)*(num_pg-1));
                      free(old);
                      files[num_pg-1].path=(unsigned char *)
                        calloc(strlen(PG_PATH),sizeof(unsigned char));
                      strcpy(files[num_pg-1].path,PG_PATH);
                      files[num_pg-1].path[0]=sw_drive;
                      files[num_pg-1].name=(unsigned char *)
                        calloc(strlen(PG_NAME),sizeof(unsigned char));
                      strcpy(files[num_pg-1].name,PG_NAME);
                      files[num_pg-1].init=sw_initial;
                      files[num_pg-1].max=sw_maximum;
                      files[num_pg-1].deleted=0;
                      update=1;
                      w_data_size+=strlen(files[num_pg-1].path);
                      w_data_size+=strlen(files[num_pg-1].name);
                      w_data_size+=19;      /* Space + number */
                      w_data_size+=19;      /* Space + number */
                      w_data_size+=2;       /* Terminating null(s) */
                      break;
    case CMD_DELETE : if((!(switches&OPT_DRV))&&(!(switches&OPT_PAG)))
                        break;
                      if((switches&OPT_DRV)&&(!(switches&OPT_PAG)))
                        sw_pagefile=matchpage2drv(num_pg,sw_drive,files);
                      if(sw_pagefile==-1)
                        break;
                      files[sw_pagefile].deleted=1;
                      update=1;
                      break;
  }

  /******************************************************************
  * If the update flag has been set, write the pagefile array back
  *   out to the registry key.
  ******************************************************************/
  if(update) {
    w_data=(unsigned char *)calloc(w_data_size,sizeof(unsigned char));
    struct2wbuffer(num_pg,w_data,files);
    rtc=RegSetValueEx(key,VMM_VAL,0,REG_MULTI_SZ,w_data,w_data_size);
    if(rtc!=ERROR_SUCCESS)
      printf("Error updating value = %ld\n",rtc);
  }

  EXIT:                           /* Exit with errorlevel */
  if(errorlevel<0x01000000) {
    free(w_data);                 /* Free the write buffer */
    /****************************************************************
    * Loop thru the array of PageFile structures, and free 
    * the path and name fields.
    ****************************************************************/
    for(ctr=0; ctr<num_pg; ctr++) {
      free(files[ctr].path);
      free(files[ctr].name);
    }  
    free(files);                  /* Free the array */
    free(r_data);                 /* Free the read buffer */
    rtc=RegCloseKey(key);         /* Close the Registry */
    free(opts);                   /* Free options array */
  }
  else {
    switch(errorlevel) {
      case 0x04000000 : free(r_data);        /* Free the read buffer */
                        rtc=RegCloseKey(key);/* Close the Registry */
      case 0x05000000 : free(opts);          /* Free options array */
    }
  }
  if(update) {
    printf("\nREMEMBER: Changes will NOT take effect UNTIL\n");
    printf("the system is REBOOTED\n");
  }
  printf("\nBATCHVMM for WinNT\n");
  printf("BATCH Virtual Memory Manager\n");
  printf("By: Paul Trout, Bionic Fish Software Foundry\n");
  printf("23 Apr. 1998 - V1R0\n");
  return(errorlevel);
}

